from __future__ import annotations

from dissect.target.helpers.magic.magic import Magic, from_buffer, from_descriptor, from_entry, from_fh, from_file

__all__ = [
    "Magic",
    "from_buffer",
    "from_descriptor",
    "from_entry",
    "from_fh",
    "from_file",
]
